searchData={"items":[{"type":"module","title":"TypedStruct","doc":"TypedStruct is a library for defining structs with a type without writing\nboilerplate code.\n\n**NOTE:** This is an active fork of the original\n[typedstruct](https://github.com/ejpcmac/typed_struct) work by Jean-Philippe Cugnet, which seems\nto be no longer maintained.  This version adds type information to Erlang\nrecords and makes the project compile under OTP-26 and later.","ref":"TypedStruct.html"},{"type":"module","title":"Rationale - TypedStruct","doc":"To define a struct in Elixir, you probably want to define three things:\n\n* the struct itself, with default values,\n* the list of enforced keys,\n* its associated type.\n\nIt ends up in something like this:\n\n```elixir\ndefmodule Person do\n  @moduledoc \"\"\"\n  A struct representing a person.\n  \"\"\"\n\n  @enforce_keys [:name]\n  defstruct name: nil,\n            age: nil,\n            happy?: true,\n            phone: nil\n\n  @typedoc \"A person\"\n  @type t() :: %__MODULE__{\n          name: String.t(),\n          age: non_neg_integer() | nil,\n          happy?: boolean(),\n          phone: String.t() | nil\n        }\nend\n```\n\nIn the example above you can notice several points:\n\n* the keys are present in both the `defstruct` and type definition,\n* enforced keys must also be written in `@enforce_keys`,\n* if a key has no default value and is not enforced, its type should be\n  nullable.\n\nIf you want to add a field in the struct, you must therefore:\n\n* add the key with its default value in the `defstruct` list,\n* add the key with its type in the type definition.\n\nIf the field is not optional, you should even add it to `@enforce_keys`. This is\nway too much work for lazy people like me, and moreover it can be error-prone.\n\nIt would be way better if we could write something like this:\n\n```elixir\ndefmodule Person do\n  @moduledoc \"\"\"\n  A struct representing a person.\n  \"\"\"\n\n  use TypedStruct\n\n  typedstruct do\n    @typedoc \"A person\"\n\n    field :name, String.t(), enforce: true\n    field :age, non_neg_integer()\n    field :happy?, boolean(), default: true\n    field :phone, String.t()\n  end\nend\n```\n\nThanks to TypedStruct, this is now possible :)","ref":"TypedStruct.html#module-rationale"},{"type":"module","title":"Usage - TypedStruct","doc":"","ref":"TypedStruct.html#module-usage"},{"type":"module","title":"Setup - TypedStruct","doc":"To use TypedStruct in your project, add this to your Mix dependencies:\n\n```elixir\n{:typedstruct, \"~> 0.5\"}\n```\n\nIf you do not plan to compile modules using TypedStruct at runtime, you can add\n`runtime: false` to the dependency tuple as TypedStruct is only used at build\ntime.\n\nIf you want to avoid `mix format` putting parentheses on field definitions,\nyou can add to your `.formatter.exs`:\n\n```elixir\n[\n  ...,\n  import_deps: [:typedstruct]\n]\n```","ref":"TypedStruct.html#module-setup"},{"type":"module","title":"General usage - TypedStruct","doc":"To define a typed struct, use\n[`TypedStruct`](https://hexdocs.pm/typedstruct/TypedStruct.html), then define\nyour struct within a `typedstruct` block:\n\n```elixir\ndefmodule MyStruct do\n  # Use TypedStruct to import the typedstruct macro.\n  use TypedStruct\n\n  # Define your struct.\n  typedstruct do\n    # Define each field with the field macro.\n    field :a_string, String.t()\n\n    # You can set a default value.\n    field :string_with_default, String.t(), default: \"default\"\n\n    # You can enforce a field.\n    field :enforced_field, integer(), enforce: true\n  end\nend\n```\n\nEach field is defined through the\n[`field/2`](https://hexdocs.pm/typedstruct/TypedStruct.html#field/2) macro.\n\nTo define a record use the `typedrecord` block:\n\n```elixir\ndefmodule Person do\n  use TypedStruct\n\n  typedrecord :person do\n    @typedoc \"A person\"\n\n    field :name, String.t(),\n    field :age,  non_neg_integer(), default: 0\n  end\nend\n```","ref":"TypedStruct.html#module-general-usage"},{"type":"module","title":"Options - TypedStruct","doc":"If you want to enforce all the keys by default, you can do:\n\n```elixir\ndefmodule MyStruct do\n  use TypedStruct\n\n  # Enforce keys by default.\n  typedstruct enforce: true do\n    # This key is enforced.\n    field :enforced_by_default, term()\n\n    # You can override the default behaviour.\n    field :not_enforced, term(), enforce: false\n\n    # A key with a default value is not enforced.\n    field :not_enforced_either, integer(), default: 1\n  end\nend\n```\nYou can also generate an opaque or private type for the struct by using\nthe `visibility: :opaque | :private | :public` option:\n\n```elixir\ndefmodule MyOpaqueStruct do\n  use TypedStruct\n\n  # Generate an opaque type for the struct.\n  typedstruct visibility: :opaque do\n    field :name, String.t()\n  end\nend\n```\n\n```elixir\ndefmodule MyPrivateStruct do\n  use TypedStruct\n\n  # Generate a private type for the struct.\n  typedstruct visibility: :private do\n    field :name, String.t()\n  end\nend\n```\n\nIf you often define submodules containing only a struct, you can avoid\nboilerplate code:\n\n```elixir\ndefmodule MyModule do\n  use TypedStruct\n\n  # You now have %MyModule.Struct{}.\n  typedstruct module: Struct do\n    field :field, term()\n  end\nend\n```","ref":"TypedStruct.html#module-options"},{"type":"module","title":"Documentation - TypedStruct","doc":"To add a `@typedoc` to the struct type, just add the attribute in the\n`typedstruct` block:\n\n```elixir\ntypedstruct do\n  @typedoc \"A typed struct\"\n\n  field :a_string, String.t()\n  field :an_int, integer()\nend\n```\n\nYou can also document submodules this way:\n\n```elixir\ntypedstruct module: MyStruct do\n  @moduledoc \"A submodule with a typed struct.\"\n  @typedoc \"A typed struct in a submodule\"\n\n  field :a_string, String.t()\n  field :an_int, integer()\nend\n```","ref":"TypedStruct.html#module-documentation"},{"type":"module","title":"Plugins - TypedStruct","doc":"It is possible to extend the scope of TypedStruct by using its plugin interface,\nas described in\n[`TypedStruct.Plugin`](https://hexdocs.pm/typedstruct/TypedStruct.Plugin.html).\nFor instance, to automatically generate lenses with the\n[Lens](https://github.com/obrok/lens) library, you can use\n[`TypedStructLens`](https://github.com/saleyn/typedstruct_lens) and do:\n\n```elixir\ndefmodule MyStruct do\n  use TypedStruct\n\n  typedstruct do\n    plugin TypedStructLens\n\n    field :a_field, String.t()\n    field :other_field, atom()\n  end\n\n  @spec change(t()) :: t()\n  def change(data) do\n    # a_field/0 is generated by TypedStructLens.\n    lens = a_field()\n    put_in(data, [lens], \"Changed\")\n  end\nend\n```\n\nPresently plugins are not supported by the `typedrecord` block.","ref":"TypedStruct.html#module-plugins"},{"type":"module","title":"Some available plugins - TypedStruct","doc":"* [`typedstruct_lens`](https://github.com/saleyn/typedstruct_lens) –\n    Integration with the [Lens](https://github.com/obrok/lens) library.\n* [`typedstruct_legacy_reflection`](https://github.com/saleyn/typedstruct_legacy_reflection)\n  – Re-enables the legacy reflection functions from TypedStruct 0.1.x.\n* [`typed_struct_ctor`](https://hexdocs.pm/typed_struct_ctor/TypedStructCtor.html)\n  – Adds validating (`new` and `from`) constructor functions.\n  > Try the macro out in real time without having to install or write any of your own code.  \n  > All you need is a running instance of [Livebook](https://livebook.dev/)\n  > \n  > [![Run in Livebook](https://livebook.dev/badge/v1/blue.svg)](https://livebook.dev/run?url=https://github.com/withbelay/typed_struct_ctor/blob/main/try_it_out.livemd)\n\nThis list is not meant to be exhaustive, please [search for “typedstruct” on\nhex.pm](https://hex.pm/packages?search=typedstruct) for other results. If you\nwant your plugin to appear here, please open an issue.","ref":"TypedStruct.html#module-some-available-plugins"},{"type":"module","title":"What do I get? - TypedStruct","doc":"When defining an empty `typedstruct` block:\n\n```elixir\ndefmodule Example do\n  use TypedStruct\n\n  typedstruct do\n  end\nend\n```\n\nyou get an empty struct with its module type `t()`:\n\n```elixir\ndefmodule Example do\n  @enforce_keys []\n  defstruct []\n\n  @type t() :: %__MODULE__{}\nend\n```\n\nEach `field` call adds information to the struct, `@enforce_keys` and the type\n`t()`.\n\nA field with no options adds the name to the `defstruct` list, with `nil` as\ndefault. The type itself is made nullable:\n\n```elixir\ndefmodule Example do\n  use TypedStruct\n\n  typedstruct do\n    field :name, String.t()\n  end\nend\n```\n\nbecomes:\n\n```elixir\ndefmodule Example do\n  @enforce_keys []\n  defstruct name: nil\n\n  @type t() :: %__MODULE__{\n          name: String.t() | nil\n        }\nend\n```\n\nThe `default` option adds the default value to the `defstruct`:\n\n```elixir\nfield :name, String.t(), default: \"John Smith\"\n\n# Becomes\ndefstruct name: \"John Smith\"\n```\n\nWhen set to `true`, the `enforce` option enforces the key by adding it to the\n`@enforce_keys` attribute.\n\n```elixir\nfield :name, String.t(), enforce: true\n\n# Becomes\n@enforce_keys [:name]\ndefstruct name: nil\n```\n\nIn both cases, the type has no reason to be nullable anymore by default. In one\ncase the field is filled with its default value and not `nil`, and in the other\ncase it is enforced. Both options would generate the following type:\n\n```elixir\n@type t() :: %__MODULE__{\n        name: String.t() # Not nullable\n      }\n```\n\nPassing `opaque: true` replaces `@type` with `@opaque` in the struct type\nspecification:\n\n```elixir\ntypedstruct opaque: true do\n  field :name, String.t()\nend\n```\n\ngenerates the following type:\n\n```elixir\n@opaque t() :: %__MODULE__{\n          name: String.t()\n        }\n```\n\nWhen passing `module: ModuleName`, the whole `typedstruct` block is wrapped in a\nmodule definition. This way, the following definition:\n\n```elixir\ndefmodule MyModule do\n  use TypedStruct\n\n  typedstruct module: Struct do\n    field :field, term()\n  end\nend\n```\n\nbecomes:\n\n```elixir\ndefmodule MyModule do\n  defmodule Struct do\n    @enforce_keys []\n    defstruct field: nil\n\n    @type t() :: %__MODULE__{\n            field: term() | nil\n          }\n  end\nend\n```\n\nTo define a typed record, the following definition of the `typedrecord`:\n\n```elixir\ndefmodule Person do\n  use TypedStruct\n  typedrecord :person do\n    @typedoc \"A person\"\n    field :name, String.t()\n    field :age,  non_neg_integer(), default: 0\n  end\nend\n```\n\nbecomes:\n\n```elixir\ndefmodule Person do\n  use Record\n  Record.defrecord(:person, name: nil, age: 0)\n  @type person :: {:person, String.t()|nil, non_neg_integer()}\nend\n```","ref":"TypedStruct.html#module-what-do-i-get"},{"type":"macro","title":"TypedStruct.field/3","doc":"Defines a field in a typed struct.","ref":"TypedStruct.html#field/3"},{"type":"macro","title":"Example - TypedStruct.field/3","doc":"# A field named :example of type String.t()\n    field :example, String.t()","ref":"TypedStruct.html#field/3-example"},{"type":"macro","title":"Options - TypedStruct.field/3","doc":"* `default` - sets the default value for the field\n  * `enforce` - if set to true, enforces the field and makes its type\n    non-nullable","ref":"TypedStruct.html#field/3-options"},{"type":"macro","title":"TypedStruct.plugin/2","doc":"Registers a plugin for the currently defined struct.","ref":"TypedStruct.html#plugin/2"},{"type":"macro","title":"Example - TypedStruct.plugin/2","doc":"typedstruct do\n      plugin MyPlugin\n\n      field :a_field, String.t()\n    end\n\nFor more information on how to define your own plugins, please see\n`TypedStruct.Plugin`. To use a third-party plugin, please refer directly to\nits documentation.","ref":"TypedStruct.html#plugin/2-example"},{"type":"macro","title":"TypedStruct.typedrecord/3","doc":"Defines a typed record.\n\nInside a `typedrecord` block, each field is defined through the `field/2`\nmacro.","ref":"TypedStruct.html#typedrecord/3"},{"type":"macro","title":"Options - TypedStruct.typedrecord/3","doc":"* `tag` - if set, used as the `tag` parameter passed to `Record.defrecord/3`.\n  * `visibility` - one of the values: `:public` (default), `:private`, `:opaque`.\n  * `module` - if set, creates the struct in a submodule named `module`.","ref":"TypedStruct.html#typedrecord/3-options"},{"type":"macro","title":"TypedStruct.typedstruct/2","doc":"Defines a typed struct.\n\nInside a `typedstruct` block, each field is defined through the `field/2`\nmacro.","ref":"TypedStruct.html#typedstruct/2"},{"type":"macro","title":"Options - TypedStruct.typedstruct/2","doc":"* `enforce` - if set to true, sets `enforce: true` to all fields by default.\n    This can be overridden by setting `enforce: false` or a default value on\n    individual fields.\n  * `visibility` - one of the values: `:public` (default), `:private`, `:opaque`.\n  * `module` - if set, creates the struct in a submodule named `module`.\n  * `opaque` - (deprecated) if set to true, creates an opaque type for the struct.","ref":"TypedStruct.html#typedstruct/2-options"},{"type":"macro","title":"Examples - TypedStruct.typedstruct/2","doc":"defmodule MyStruct do\n      use TypedStruct\n\n      typedstruct do\n        field :field_one, String.t()\n        field :field_two, integer(), enforce: true\n        field :field_three, boolean(), enforce: true\n        field :field_four, atom(), default: :hey\n      end\n    end\n\nThe following is an equivalent using the *enforce by default* behaviour:\n\n    defmodule MyStruct do\n      use TypedStruct\n\n      typedstruct enforce: true do\n        field :field_one, String.t(), enforce: false\n        field :field_two, integer()\n        field :field_three, boolean()\n        field :field_four, atom(), default: :hey\n      end\n    end\n\nYou can create the struct in a submodule instead:\n\n    defmodule MyModule do\n      use TypedStruct\n\n      typedstruct module: Struct do\n        field :field_one, String.t()\n        field :field_two, integer(), enforce: true\n        field :field_three, boolean(), enforce: true\n        field :field_four, atom(), default: :hey\n      end\n    end","ref":"TypedStruct.html#typedstruct/2-examples"},{"type":"behaviour","title":"TypedStruct.Plugin","doc":"This module defines the plugin interface for TypedStruct.","ref":"TypedStruct.Plugin.html"},{"type":"behaviour","title":"Rationale - TypedStruct.Plugin","doc":"Sometimes you may want to define helpers on your structs, for all their fields\nor for the struct as a whole. This plugin interface lets you integrate your\nown needs with TypedStruct.","ref":"TypedStruct.Plugin.html#module-rationale"},{"type":"behaviour","title":"Plugin definition - TypedStruct.Plugin","doc":"A TypedStruct plugin is a module that implements `TypedStruct.Plugin`. Three\ncallbacks are available to you for injecting code at different steps:\n\n  * `c:init/1` lets you inject code where the `TypedStruct.plugin/2` macro is\n    called,\n  * `c:field/4` lets you inject code on each field definition,\n  * `c:after_definition/1` lets you insert code after the struct and its type\n    have been defined.\n\n`use`-ing this module will inject default implementations of all\nthree, so you only have to implement those you care about.","ref":"TypedStruct.Plugin.html#module-plugin-definition"},{"type":"behaviour","title":"Example - TypedStruct.Plugin","doc":"As an example, let’s define a plugin that allows users to add an optional\ndescription to their structs and fields. This plugin also takes an `upcase`\noption. If set to `true`, all the descriptions are then upcased. It would be\nused this way:\n\n    defmodule MyStruct do\n      use TypedStruct\n\n      typedstruct do\n        # We import the plugin with the upcase option set to `true`.\n        plugin DescribedStruct, upcase: true\n\n        # We can now set a description for the struct.\n        description \"My struct\"\n\n        # We can also set a description on a field.\n        field :a_field, String.t(), description: \"A field\"\n        field :second_field, boolean()\n      end\n    end\n\nOnce compiled, we would optain:\n\n    iex> MyStruct.struct_description()\n    \"MY STRUCT\"\n    iex> MyStruct.field_description(:a_field)\n    \"A FIELD\"\n\nFollows the plugin definition:\n\n    defmodule DescribedStruct do\n      use TypedStruct.Plugin\n\n      # The init macro lets you inject code where the plugin macro is called.\n      # You can think a bit of it like a `use` but for the scope of the\n      # typedstruct block.\n      @impl true\n      @spec init(keyword()) :: Macro.t()\n      defmacro init(opts) do\n        quote do\n          # Let’s import our custom `description` macro defined below so our\n          # users can use it when defining their structs.\n          import TypedStructDemoPlugin, only: [description: 1]\n\n          # Let’s also store the upcase option in an attribute so we can\n          # access it from the code injected by our `description/1` macro.\n          @upcase unquote(opts)[:upcase]\n        end\n      end\n\n      # This is a public macro our users can call in their typedstruct blocks.\n      @spec description(String.t()) :: Macro.t()\n      defmacro description(description) do\n        quote do\n          # Here we simply evaluate the result of __description__/2. We need\n          # this indirection to be able to use @upcase after is has been\n          # evaluated, but still in the code generation process. This way, we\n          # can upcase the strings *at build time* if needed. It’s just a tiny\n          # refinement :-)\n          Module.eval_quoted(\n            __MODULE__,\n            TypedStructDemoPlugin.__description__(__MODULE__, unquote(description))\n          )\n        end\n      end\n\n      @spec __description__(module(), String.t()) :: Macro.t()\n      def __description__(module, description) do\n        # Maybe upcase the description at build time.\n        description =\n          module\n          |> Module.get_attribute(:upcase)\n          |> maybe_upcase(description)\n\n        quote do\n          # Let’s just generate a constant function that returns the\n          # description.\n          def struct_description, do: unquote(description)\n        end\n      end\n\n      # The field callback is called for each field defined in the typedstruct\n      # block. You get exactly what the user has passed to the field macro,\n      # plus options from every plugin init. The `env` variable contains the\n      # environment as it stood at the moment of the corresponding\n      # `TypedStruct.field/3` call.\n      @impl true\n      @spec field(atom(), any(), keyword(), Macro.Env.t()) :: Macro.t()\n      def field(name, _type, opts, _env) do\n        # Same as for the struct description, we want to upcase at build time\n        # if necessary. As we do not have access to the module here, we cannot\n        # access @upcase. This is not an issue since the option is\n        # automatically added to `opts`, in addition to the options passed to\n        # the field macro.\n        description = maybe_upcase(opts[:upcase], opts[:description] || \"\")\n\n        quote do\n          # We define a clause matching the field name returning its optional\n          # description.\n          def field_description(unquote(name)), do: unquote(description)\n        end\n      end\n\n      defp maybe_upcase(true, description), do: String.upcase(description)\n      defp maybe_upcase(_, description), do: description\n\n      # The after_definition callback is called after the struct and its type\n      # have been defined, at the end of the `typedstruct` block.\n      @impl true\n      @spec after_definition(opts :: keyword()) :: Macro.t()\n      def after_definition(_opts) do\n        quote do\n          # Here we just clean the @upcase attribute so that it does not\n          # pollute our user’s modules.\n          Module.delete_attribute(__MODULE__, :upcase)\n        end\n      end\n    end","ref":"TypedStruct.Plugin.html#module-example"},{"type":"callback","title":"TypedStruct.Plugin.after_definition/1","doc":"Injects code after the struct and its type have been defined.","ref":"TypedStruct.Plugin.html#c:after_definition/1"},{"type":"callback","title":"TypedStruct.Plugin.field/3","doc":"","ref":"TypedStruct.Plugin.html#c:field/3"},{"type":"callback","title":"TypedStruct.Plugin.field/4","doc":"Injects code after each field definition.\n\n`name` and `type` are the exact values passed to the `TypedStruct.field/3`\nmacro in the `typedstruct` block. `opts` is the concatenation of the options\npassed to the `field` macro and those from the plugin init. `env` is the\nenvironment at the time of each field definition.","ref":"TypedStruct.Plugin.html#c:field/4"},{"type":"macrocallback","title":"TypedStruct.Plugin.init/1","doc":"Injects code where `TypedStruct.plugin/2` is called.","ref":"TypedStruct.Plugin.html#c:init/1"},{"type":"extras","title":"Overview","doc":"# TypedStruct\n\n[![Build Status](https://github.com/saleyn/typedstruct/actions/workflows/ci.yml/badge.svg)](https://github.com/saleyn/typedstruct/actions/workflows/ci.yml)\n[![hex.pm version](https://img.shields.io/hexpm/v/typedstruct.svg?style=flat)](https://hex.pm/packages/typedstruct)\n[![Hex Docs](https://img.shields.io/badge/hex-docs-lightgreen.svg?style=flat)](https://hexdocs.pm/typedstruct/)\n[![License](https://img.shields.io/hexpm/l/typedstruct.svg?style=flat)](https://github.com/saleyn/typedstruct/blob/main/LICENSE.md)\n\n<!-- @moduledoc -->\n\nTypedStruct is a library for defining structs with a type without writing\nboilerplate code.\n\n**NOTE:** This is an active fork of the original\n[typedstruct](https://github.com/ejpcmac/typed_struct) work by Jean-Philippe Cugnet, which seems\nto be no longer maintained.  This version adds type information to Erlang\nrecords and makes the project compile under OTP-26 and later.","ref":"readme.html"},{"type":"extras","title":"Rationale - Overview","doc":"To define a struct in Elixir, you probably want to define three things:\n\n* the struct itself, with default values,\n* the list of enforced keys,\n* its associated type.\n\nIt ends up in something like this:\n\n```elixir\ndefmodule Person do\n  @moduledoc \"\"\"\n  A struct representing a person.\n  \"\"\"\n\n  @enforce_keys [:name]\n  defstruct name: nil,\n            age: nil,\n            happy?: true,\n            phone: nil\n\n  @typedoc \"A person\"\n  @type t() :: %__MODULE__{\n          name: String.t(),\n          age: non_neg_integer() | nil,\n          happy?: boolean(),\n          phone: String.t() | nil\n        }\nend\n```\n\nIn the example above you can notice several points:\n\n* the keys are present in both the `defstruct` and type definition,\n* enforced keys must also be written in `@enforce_keys`,\n* if a key has no default value and is not enforced, its type should be\n  nullable.\n\nIf you want to add a field in the struct, you must therefore:\n\n* add the key with its default value in the `defstruct` list,\n* add the key with its type in the type definition.\n\nIf the field is not optional, you should even add it to `@enforce_keys`. This is\nway too much work for lazy people like me, and moreover it can be error-prone.\n\nIt would be way better if we could write something like this:\n\n```elixir\ndefmodule Person do\n  @moduledoc \"\"\"\n  A struct representing a person.\n  \"\"\"\n\n  use TypedStruct\n\n  typedstruct do\n    @typedoc \"A person\"\n\n    field :name, String.t(), enforce: true\n    field :age, non_neg_integer()\n    field :happy?, boolean(), default: true\n    field :phone, String.t()\n  end\nend\n```\n\nThanks to TypedStruct, this is now possible :)","ref":"readme.html#rationale"},{"type":"extras","title":"Usage - Overview","doc":"","ref":"readme.html#usage"},{"type":"extras","title":"Setup - Overview","doc":"To use TypedStruct in your project, add this to your Mix dependencies:\n\n```elixir\n{:typedstruct, \"~> 0.5\"}\n```\n\nIf you do not plan to compile modules using TypedStruct at runtime, you can add\n`runtime: false` to the dependency tuple as TypedStruct is only used at build\ntime.\n\nIf you want to avoid `mix format` putting parentheses on field definitions,\nyou can add to your `.formatter.exs`:\n\n```elixir\n[\n  ...,\n  import_deps: [:typedstruct]\n]\n```","ref":"readme.html#setup"},{"type":"extras","title":"General usage - Overview","doc":"To define a typed struct, use\n[`TypedStruct`](https://hexdocs.pm/typedstruct/TypedStruct.html), then define\nyour struct within a `typedstruct` block:\n\n```elixir\ndefmodule MyStruct do\n  # Use TypedStruct to import the typedstruct macro.\n  use TypedStruct\n\n  # Define your struct.\n  typedstruct do\n    # Define each field with the field macro.\n    field :a_string, String.t()\n\n    # You can set a default value.\n    field :string_with_default, String.t(), default: \"default\"\n\n    # You can enforce a field.\n    field :enforced_field, integer(), enforce: true\n  end\nend\n```\n\nEach field is defined through the\n[`field/2`](https://hexdocs.pm/typedstruct/TypedStruct.html#field/2) macro.\n\nTo define a record use the `typedrecord` block:\n\n```elixir\ndefmodule Person do\n  use TypedStruct\n\n  typedrecord :person do\n    @typedoc \"A person\"\n\n    field :name, String.t(),\n    field :age,  non_neg_integer(), default: 0\n  end\nend\n```","ref":"readme.html#general-usage"},{"type":"extras","title":"Options - Overview","doc":"If you want to enforce all the keys by default, you can do:\n\n```elixir\ndefmodule MyStruct do\n  use TypedStruct\n\n  # Enforce keys by default.\n  typedstruct enforce: true do\n    # This key is enforced.\n    field :enforced_by_default, term()\n\n    # You can override the default behaviour.\n    field :not_enforced, term(), enforce: false\n\n    # A key with a default value is not enforced.\n    field :not_enforced_either, integer(), default: 1\n  end\nend\n```\nYou can also generate an opaque or private type for the struct by using\nthe `visibility: :opaque | :private | :public` option:\n\n```elixir\ndefmodule MyOpaqueStruct do\n  use TypedStruct\n\n  # Generate an opaque type for the struct.\n  typedstruct visibility: :opaque do\n    field :name, String.t()\n  end\nend\n```\n\n```elixir\ndefmodule MyPrivateStruct do\n  use TypedStruct\n\n  # Generate a private type for the struct.\n  typedstruct visibility: :private do\n    field :name, String.t()\n  end\nend\n```\n\nIf you often define submodules containing only a struct, you can avoid\nboilerplate code:\n\n```elixir\ndefmodule MyModule do\n  use TypedStruct\n\n  # You now have %MyModule.Struct{}.\n  typedstruct module: Struct do\n    field :field, term()\n  end\nend\n```","ref":"readme.html#options"},{"type":"extras","title":"Documentation - Overview","doc":"To add a `@typedoc` to the struct type, just add the attribute in the\n`typedstruct` block:\n\n```elixir\ntypedstruct do\n  @typedoc \"A typed struct\"\n\n  field :a_string, String.t()\n  field :an_int, integer()\nend\n```\n\nYou can also document submodules this way:\n\n```elixir\ntypedstruct module: MyStruct do\n  @moduledoc \"A submodule with a typed struct.\"\n  @typedoc \"A typed struct in a submodule\"\n\n  field :a_string, String.t()\n  field :an_int, integer()\nend\n```","ref":"readme.html#documentation"},{"type":"extras","title":"Plugins - Overview","doc":"It is possible to extend the scope of TypedStruct by using its plugin interface,\nas described in\n[`TypedStruct.Plugin`](https://hexdocs.pm/typedstruct/TypedStruct.Plugin.html).\nFor instance, to automatically generate lenses with the\n[Lens](https://github.com/obrok/lens) library, you can use\n[`TypedStructLens`](https://github.com/saleyn/typedstruct_lens) and do:\n\n```elixir\ndefmodule MyStruct do\n  use TypedStruct\n\n  typedstruct do\n    plugin TypedStructLens\n\n    field :a_field, String.t()\n    field :other_field, atom()\n  end\n\n  @spec change(t()) :: t()\n  def change(data) do\n    # a_field/0 is generated by TypedStructLens.\n    lens = a_field()\n    put_in(data, [lens], \"Changed\")\n  end\nend\n```\n\nPresently plugins are not supported by the `typedrecord` block.","ref":"readme.html#plugins"},{"type":"extras","title":"Some available plugins - Overview","doc":"* [`typedstruct_lens`](https://github.com/saleyn/typedstruct_lens) –\n    Integration with the [Lens](https://github.com/obrok/lens) library.\n* [`typedstruct_legacy_reflection`](https://github.com/saleyn/typedstruct_legacy_reflection)\n  – Re-enables the legacy reflection functions from TypedStruct 0.1.x.\n* [`typed_struct_ctor`](https://hexdocs.pm/typed_struct_ctor/TypedStructCtor.html)\n  – Adds validating (`new` and `from`) constructor functions.\n  > Try the macro out in real time without having to install or write any of your own code.  \n  > All you need is a running instance of [Livebook](https://livebook.dev/)\n  > \n  > [![Run in Livebook](https://livebook.dev/badge/v1/blue.svg)](https://livebook.dev/run?url=https://github.com/withbelay/typed_struct_ctor/blob/main/try_it_out.livemd)\n\nThis list is not meant to be exhaustive, please [search for “typedstruct” on\nhex.pm](https://hex.pm/packages?search=typedstruct) for other results. If you\nwant your plugin to appear here, please open an issue.","ref":"readme.html#some-available-plugins"},{"type":"extras","title":"What do I get? - Overview","doc":"When defining an empty `typedstruct` block:\n\n```elixir\ndefmodule Example do\n  use TypedStruct\n\n  typedstruct do\n  end\nend\n```\n\nyou get an empty struct with its module type `t()`:\n\n```elixir\ndefmodule Example do\n  @enforce_keys []\n  defstruct []\n\n  @type t() :: %__MODULE__{}\nend\n```\n\nEach `field` call adds information to the struct, `@enforce_keys` and the type\n`t()`.\n\nA field with no options adds the name to the `defstruct` list, with `nil` as\ndefault. The type itself is made nullable:\n\n```elixir\ndefmodule Example do\n  use TypedStruct\n\n  typedstruct do\n    field :name, String.t()\n  end\nend\n```\n\nbecomes:\n\n```elixir\ndefmodule Example do\n  @enforce_keys []\n  defstruct name: nil\n\n  @type t() :: %__MODULE__{\n          name: String.t() | nil\n        }\nend\n```\n\nThe `default` option adds the default value to the `defstruct`:\n\n```elixir\nfield :name, String.t(), default: \"John Smith\"\n\n# Becomes\ndefstruct name: \"John Smith\"\n```\n\nWhen set to `true`, the `enforce` option enforces the key by adding it to the\n`@enforce_keys` attribute.\n\n```elixir\nfield :name, String.t(), enforce: true\n\n# Becomes\n@enforce_keys [:name]\ndefstruct name: nil\n```\n\nIn both cases, the type has no reason to be nullable anymore by default. In one\ncase the field is filled with its default value and not `nil`, and in the other\ncase it is enforced. Both options would generate the following type:\n\n```elixir\n@type t() :: %__MODULE__{\n        name: String.t() # Not nullable\n      }\n```\n\nPassing `opaque: true` replaces `@type` with `@opaque` in the struct type\nspecification:\n\n```elixir\ntypedstruct opaque: true do\n  field :name, String.t()\nend\n```\n\ngenerates the following type:\n\n```elixir\n@opaque t() :: %__MODULE__{\n          name: String.t()\n        }\n```\n\nWhen passing `module: ModuleName`, the whole `typedstruct` block is wrapped in a\nmodule definition. This way, the following definition:\n\n```elixir\ndefmodule MyModule do\n  use TypedStruct\n\n  typedstruct module: Struct do\n    field :field, term()\n  end\nend\n```\n\nbecomes:\n\n```elixir\ndefmodule MyModule do\n  defmodule Struct do\n    @enforce_keys []\n    defstruct field: nil\n\n    @type t() :: %__MODULE__{\n            field: term() | nil\n          }\n  end\nend\n```\n\nTo define a typed record, the following definition of the `typedrecord`:\n\n```elixir\ndefmodule Person do\n  use TypedStruct\n  typedrecord :person do\n    @typedoc \"A person\"\n    field :name, String.t()\n    field :age,  non_neg_integer(), default: 0\n  end\nend\n```\n\nbecomes:\n\n```elixir\ndefmodule Person do\n  use Record\n  Record.defrecord(:person, name: nil, age: 0)\n  @type person :: {:person, String.t()|nil, non_neg_integer()}\nend\n```\n\n<!-- @moduledoc -->","ref":"readme.html#what-do-i-get"},{"type":"extras","title":"Initial roadmap - Overview","doc":"* [x] Struct definition\n* [x] Type definition (with nullable types)\n* [x] Default values\n* [x] Enforced keys (non-nullable types)\n* [x] Plugin API","ref":"readme.html#initial-roadmap"},{"type":"extras","title":"Plugin ideas - Overview","doc":"* [ ] Default value type-checking (is it possible?)\n* [ ] Guard generation\n* [x] Integration with [Lens](https://github.com/obrok/lens)\n* [ ] Integration with [Ecto](https://github.com/elixir-ecto/ecto)","ref":"readme.html#plugin-ideas"},{"type":"extras","title":"Related libraries - Overview","doc":"* [Domo](https://github.com/IvanRublev/Domo): a library to validate structs that\n    define a `t()` type, like the one generated by `TypedStruct`.\n* [TypedEctoSchema](https://github.com/bamorim/typed_ecto_schema): a library\n    that provides a DSL on top of `Ecto.Schema` to achieve the same result as\n    `TypedStruct`, with `Ecto`.\n\n## [Contributing](CONTRIBUTING.md)\n\nBefore contributing to this project, please read the\n[CONTRIBUTING.md](CONTRIBUTING.md).","ref":"readme.html#related-libraries"},{"type":"extras","title":"License - Overview","doc":"Copyright © 2018-2022 Jean-Philippe Cugnet and Contributors\n\nThis project is licensed under the [MIT license](./LICENSE.md).","ref":"readme.html#license"},{"type":"extras","title":"Changelog","doc":"# Changelog\n\nAll notable changes to this project will be documented in this file.\n\nThe format is based on [Keep a Changelog](https://keepachangelog.com/en/1.0.0/),\nand this project adheres to [Semantic\nVersioning](https://semver.org/spec/v2.0.0.html).\n\n## [Unreleased]\n\n## [0.3.0] - 2022-02-15","ref":"changelog.html"},{"type":"extras","title":"Added - Changelog","doc":"* [Plugin] Add the `field/4` callback with an additional `env` argument. This is\n    the same as `field/3` but gives access to the environment of the field\n    definition.","ref":"changelog.html#added"},{"type":"extras","title":"Deprecated - Changelog","doc":"* [Plugin] Deprecate the `field/3` callback in favour of `field/4`. You should\n    migrate to `field/4`, simply by adding `_env` as the last argument in your\n    implementation. Currently, if a plugin implements `field/3`, `field/4` is\n    derived from it and a compilation warning is emitted. `field/3` will be\n    removed in TypedStruct 1.0.0.","ref":"changelog.html#deprecated"},{"type":"extras","title":"Removed - Changelog","doc":"* Drop support for unsupported Elixir versions in the tests and CI. The\n    library may still be compatible, but this is not tested.","ref":"changelog.html#removed"},{"type":"extras","title":"Fixed - Changelog","doc":"* Fix the lexical scope of the `typestruct` block, so it covers it completely.\n    Previously, anything defined inside the `typedstruct` block, such as\n    aliases, would not be available for the field definitions. See #22 and #21\n    for details.\n* Fix a typo in the documentation.\n\n## [0.2.1] - 2020-07-19","ref":"changelog.html#fixed"},{"type":"extras","title":"Added - Changelog","doc":"* Add the `module: ModuleName` top-level option to create the typed struct in a\n    submodule.","ref":"changelog.html#added"},{"type":"extras","title":"Changed - Changelog","doc":"* Update the `@typedoc` example in the documentation to put it inside the\n    `typedstruct` block and not above. While putting it above works in the\n    general case, it is mandatory to put it inside the block when defining a\n    submodule.\n\n## [0.2.0] - 2020-05-31","ref":"changelog.html#changed"},{"type":"extras","title":"Added - Changelog","doc":"* Add a plugin API.","ref":"changelog.html#added"},{"type":"extras","title":"Removed - Changelog","doc":"* Remove reflection support through the `__keys__/0`, `__defaults__/0` and\n    `__types__/0` function which where defined by TypedStruct in the user\n    modules. If you rely on these functions, please use the\n    [TypedStructLegacyReflection](https://github.com/ejpcmac/typed_struct_legacy_reflection)\n    plugin to enable them again, and consider creating a plugin for your needs.","ref":"changelog.html#removed"},{"type":"extras","title":"Fixed - Changelog","doc":"* Do not enforce fields with a default value set to nil (fixes #14).\n* Prefix all internal module attributes and clean them after use (fixes #15).\n* Create a scope in the `typedstruct` block to avoid import leaks.\n\n## [0.1.4] - 2018-11-13","ref":"changelog.html#fixed"},{"type":"extras","title":"Added - Changelog","doc":"* Add the ability to generate an opaque type (#10).\n\n## [0.1.3] - 2018-09-06","ref":"changelog.html#added"},{"type":"extras","title":"Fixed - Changelog","doc":"* Fix a bug where boolean fields with `default: false` where still enforced when\n    setting `enforce: true` at top-level.\n\n## [0.1.2] - 2018-09-06","ref":"changelog.html#fixed"},{"type":"extras","title":"Added - Changelog","doc":"* Add the ability to enforce keys by default (#6).","ref":"changelog.html#added"},{"type":"extras","title":"Fixed - Changelog","doc":"* Clarify the documentation about `runtime: false`.\n\n## [0.1.1] - 2018-06-20","ref":"changelog.html#fixed"},{"type":"extras","title":"Fixed - Changelog","doc":"* Do not make the type nullable when there is a default value.\n\n## [0.1.0] - 2018-06-19","ref":"changelog.html#fixed"},{"type":"extras","title":"Added - Changelog","doc":"* Struct definition\n* Type definition\n* Default values\n* Enforced keys\n\n[Unreleased]: https://github.com/saleyn/typedstruct/compare/main...develop\n[0.5.0]: https://github.com/saleyn/typedstruct/compare/v0.3.0...v0.5.0\n[0.3.0]: https://github.com/ejpcmac/typed_struct/compare/v0.2.1...v0.3.0\n[0.2.1]: https://github.com/ejpcmac/typed_struct/compare/v0.2.0...v0.2.1\n[0.2.0]: https://github.com/ejpcmac/typed_struct/compare/v0.1.4...v0.2.0\n[0.1.4]: https://github.com/ejpcmac/typed_struct/compare/v0.1.3...v0.1.4\n[0.1.3]: https://github.com/ejpcmac/typed_struct/compare/v0.1.2...v0.1.3\n[0.1.2]: https://github.com/ejpcmac/typed_struct/compare/v0.1.1...v0.1.2\n[0.1.1]: https://github.com/ejpcmac/typed_struct/compare/v0.1.0...v0.1.1\n[0.1.0]: https://github.com/ejpcmac/typed_struct/releases/tag/v0.1.0","ref":"changelog.html#added"},{"type":"extras","title":"Contributing","doc":"# Contributing to TypedStruct\n\nTypedStruct is written in [Elixir](https://elixir-lang.org).\n\nFor branching management, this project uses\n[git-flow](https://github.com/petervanderdoes/gitflow-avh). The `main` branch is\nreserved for releases: the development process occurs on `develop` and feature\nbranches. **Please never commit to `main`.**\n\nYou can easily set up a development environment featuring all the dependencies,\nincluding Elixir and `git-flow`, by using [Nix](https://nixos.org/nix/). This is\ndetailed below.","ref":"contributing.html"},{"type":"extras","title":"Setup - Contributing","doc":"","ref":"contributing.html#setup"},{"type":"extras","title":"Local repository - Contributing","doc":"1. Fork the repository.\n\n2. Clone your fork to a local repository:\n\n        $ git clone https://github.com/you/typedstruct.git\n        $ cd typedstruct\n\n3. Add the main repository as a remote:\n\n        $ git remote add upstream https://github.com/ejpcmac/typedstruct.git\n\n4. Checkout `develop`:\n\n        $ git checkout develop","ref":"contributing.html#local-repository"},{"type":"extras","title":"Development environment (without Nix) - Contributing","doc":"Install an Elixir environment, and optionally install `git-flow`.","ref":"contributing.html#development-environment-without-nix"},{"type":"extras","title":"Development environment (with Nix) - Contributing","doc":"1. Install Nix by running the script and following the instructions:\n\n        $ curl https://nixos.org/nix/install | sh\n\n2. Optionally install [direnv](https://github.com/direnv/direnv) to\n    automatically setup the environment when you enter the project directory:\n\n        $ nix-env -i direnv\n\n    In this case, you also need to add to your `~/. rc`:\n\n    ```sh\n    eval \"$(direnv hook  )\"\n    ```\n\n    *Make sure to replace ` ` by your shell, namely `bash`, `zsh`, …*\n\n3. In the project directory, if you **did not** install direnv, start a Nix\n   shell:\n\n        $ cd typedstruct\n        $ nix-shell\n\n    If you opted to use direnv, please allow the `.envrc` instead of running a\n    Nix shell manually:\n\n        $ cd typedstruct\n        $ direnv allow\n\n    In this case, direnv will automatically update your environment to behave\n    like a Nix shell whenever you enter the project directory.","ref":"contributing.html#development-environment-with-nix"},{"type":"extras","title":"Git-flow - Contributing","doc":"If you want to use `git-flow` and use the standard project configuration, please\nrun:\n\n    $ ./.gitsetup","ref":"contributing.html#git-flow"},{"type":"extras","title":"Building the project - Contributing","doc":"1. Fetch the project dependencies:\n\n        $ cd typedstruct\n        $ mix deps.get\n\n2. Run the static analyzers:\n\n        $ mix check\n\nAll the tests should pass.","ref":"contributing.html#building-the-project"},{"type":"extras","title":"Workflow - Contributing","doc":"To make a change, please use this workflow:\n\n1. Checkout `develop` and apply the last upstream changes (use rebase, not\n    merge!):\n\n        $ git checkout develop\n        $ git fetch --all --prune\n        $ git rebase upstream/develop\n\n2. For a tiny patch, create a new branch with an explicit name:\n\n        $ git checkout -b  \n\n    Alternatively, if you are working on a feature which would need more work,\n    you can create a feature branch with `git-flow`:\n\n        $ git flow feature start  \n\n    *Note: always open an issue and ask before starting a big feature, to avoid\n    it not beeing merged and your time lost.*\n\n3. Work on your feature (don’t forget to write typespecs and tests; you can\n    check your coverage with `mix coveralls.html` and open\n    `cover/excoveralls.html`):\n\n        # Some work\n        $ git commit -am \"feat: my first change\"\n        # Some work\n        $ git commit -am \"refactor: my second change\"\n        ...\n\n4. When your feature is ready, feel free to use\n    [interactive rebase](https://help.github.com/articles/about-git-rebase/) so\n    your history looks clean and is easy to follow. Then, apply the last\n    upstream changes on `develop` to prepare integration:\n\n        $ git checkout develop\n        $ git fetch --all --prune\n        $ git rebase upstream/develop\n\n5. If there were commits on `develop` since the beginning of your feature\n    branch, integrate them by **rebasing** if your branch has few commits, or\n    merging if you had a long-lived branch:\n\n        $ git checkout  \n        $ git rebase develop\n\n    *Note: the only case you should merge is when you are working on a big\n    feature. If it is the case, we should have discussed this before as stated\n    above.*\n\n6. Run the tests and static analyzers to ensure there is no regression and all\n    works as expected:\n\n        $ mix check\n\n7. If it’s all good, open a pull request to merge your branch into the `develop`\n    branch on the main repository.","ref":"contributing.html#workflow"},{"type":"extras","title":"Coding style - Contributing","doc":"Please format your code with `mix format` or your editor and follow\n[this style guide](https://github.com/christopheradams/elixir_style_guide).\n\nAll contributed code must be documented and functions must have typespecs. In\ngeneral, take your inspiration from the existing code.\n\nPlease name your commits using [Conventional\nCommits](https://www.conventionalcommits.org/en/v1.0.0/).","ref":"contributing.html#coding-style"},{"type":"extras","title":"License","doc":"# MIT License\n\nCopyright © 2018-2022 Jean-Philippe Cugnet and Contributors\n\nPermission is hereby granted, free of charge, to any person obtaining a copy\nof this software and associated documentation files (the \"Software\"), to deal\nin the Software without restriction, including without limitation the rights\nto use, copy, modify, merge, publish, distribute, sublicense, and/or sell\ncopies of the Software, and to permit persons to whom the Software is\nfurnished to do so, subject to the following conditions:\n\nThe above copyright notice and this permission notice shall be included in all\ncopies or substantial portions of the Software.\n\nTHE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\nIMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\nFITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE\nAUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER\nLIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,\nOUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE\nSOFTWARE.","ref":"license.html"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,48]}}